from enum import Enum
import logging
from typing import List, Optional
import typer
import os
import time

from typing_extensions import Annotated

from aperturedb.Query import ObjectType
from aperturedb.cli.console import console


from tqdm import tqdm

logger = logging.getLogger(__file__)
app = typer.Typer()


IngestType = Enum('IngestType', {k: str(k) for k in ObjectType._member_names_})

# This list needs to be updated when new transformers are added
# This enables CLI params.


class TransformerType(str, Enum):
    common_properties = "common_properties"
    image_properties = "image_properties"
    clip_pytorch_embeddings = "clip_pytorch_embeddings"
    facenet_pytorch_embeddings = "facenet_pytorch_embeddings"


def _debug_samples(data, sample_count, module_name):
    import json
    print(f"len(data)={len(data)}")
    if len(data) < sample_count:
        sample_count = len(data)
        console.log(
            f"Sample count is greater than data length. Using {sample_count} samples instead.")
    for i, r in enumerate(data[0:sample_count]):
        query, blobs = r
        with open(module_name + f"_{i}" + ".raw", "w") as f:
            f.write(str(query))
        with open(module_name + f"_{i}" + ".json", "w") as f:
            json.dump(query, f, indent=2, ensure_ascii=False)
        for j, blob in enumerate(blobs):
            with open(module_name + f"_{i}_{j}" + ".jpg", "wb") as f:
                f.write(blob)


def _apply_pipeline(data, transformers: List[str], **kwargs):
    pipeline = _create_pipeline(transformers)
    console.log("Applying Pipeline: \r\n" +
                "\r\n=>".join([f"{stage.__name__}[{kwargs}]" for stage in pipeline]))
    for transformer in pipeline:
        data = transformer(data, **kwargs)
        data.sample_count = data.data.sample_count
    return data


def _create_pipeline(transformers: List[str]):
    from aperturedb.transformers.common_properties import CommonProperties
    from aperturedb.transformers.image_properties import ImageProperties
    from aperturedb.transformers.clip_pytorch_embeddings import CLIPPyTorchEmbeddings
    from aperturedb.transformers.facenet_pytorch_embeddings import FacenetPyTorchEmbeddings
    from aperturedb.CommonLibrary import import_module_by_path

    # Actual collection of transformers, packaged with aperturedb.
    built_in_transformers = {
        "common_properties": CommonProperties,
        "image_properties": ImageProperties,
        "clip_pytorch_embeddings": CLIPPyTorchEmbeddings,
        "facenet_pytorch_embeddings": FacenetPyTorchEmbeddings
    }
    pipeline = []
    for transformer_name in transformers:
        transformer = None
        if transformer_name in built_in_transformers:
            transformer = built_in_transformers[transformer_name]
        else:
            try:
                transformer_module = import_module_by_path(transformer_name)
                transformer = getattr(transformer_module,
                                      transformer_module.__name__)
            except Exception as e:
                console.log(
                    f"Could not load transformer {transformer_name}: {e}")
                exit(1)
        pipeline.append(transformer)
    return pipeline


def _process_data(data, sample_count, module_name, batchsize, num_workers, stats, debug):
    if debug:
        _debug_samples(data, sample_count, module_name)
    else:
        from aperturedb.ParallelLoader import ParallelLoader
        from aperturedb.CommonLibrary import create_connector

        client = create_connector()
        loader = ParallelLoader(client)
        loader.ingest(
            data,
            stats=stats,
            batchsize=batchsize,
            numthreads=num_workers)


@app.command()
def from_generator(filepath: Annotated[str, typer.Argument(
    help="Path to python module for ingestion [BETA]")],
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
    debug: Annotated[bool, typer.Option(
        help="Debug mode")] = False,
    stats: Annotated[bool, typer.Option(
        help="Show realtime statistics with summary")] = True,
    batchsize: Annotated[int, typer.Option(
        help="Size of the batch")] = 1,
    num_workers: Annotated[int, typer.Option(
        help="Number of workers for ingestion")] = 1,
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times]")] = None,
):
    """
    Ingest data from a Data generator [BETA].
    """

    from aperturedb.CommonLibrary import import_module_by_path

    module = import_module_by_path(filepath)

    # The assumption is that the class name is same as that of the module name
    module_name  = os.path.basename(filepath)[:-3]
    mclass = getattr(module, module_name)

    # This is the dynamically loaded data generator.
    # tested with CelebADataKaggle.py, CocoDataPytorch.py and Cifar10DataTensorflow.py in examples
    start = time.time()
    data = mclass()
    data.sample_count = len(data) if sample_count == -1 else sample_count
    console.log(f"Data generator loaded in {time.time() - start} seconds")

    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{module_name}.{mclass.__name__}")

    _process_data(
        data,
        sample_count=sample_count,
        module_name=module_name,
        batchsize=batchsize,
        num_workers=num_workers,
        stats=stats,
        debug=debug
    )

    while hasattr(data, "ncalls"):
        console.log(
            f"Calls to {data}.getitem = {data.ncalls} time={data.cumulative_time}")
        data = data.data


@app.command()
def from_csv(filepath: Annotated[str, typer.Argument(
    help="Path to csv for ingestion")],
    batchsize: Annotated[int, typer.Option(
        help="Size of the batch")] = 1,
    num_workers: Annotated[int, typer.Option(
        help="Number of workers for ingestion")] = 1,
    stats: Annotated[bool, typer.Option(
        help="Show realtime statistics with summary")] = True,
    use_dask: Annotated[bool, typer.Option(
        help="Use dask based parallelization")] = False,
    ingest_type: Annotated[IngestType, typer.Option(
        help="Parser for CSV file to be used")] = IngestType.IMAGE,
    blobs_relative_to_csv: Annotated[bool, typer.Option(
        help="If true, the blob path is relative to the CSV file")] = True,
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times.]")] = None,
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
    debug: Annotated[bool, typer.Option(
        help="Debug mode")] = False,
):
    """
    Ingest data from a pre generated CSV file.
    """
    from aperturedb.ImageDataCSV import ImageDataCSV
    from aperturedb.BBoxDataCSV import BBoxDataCSV
    from aperturedb.EntityDataCSV import EntityDataCSV
    from aperturedb.BlobDataCSV import BlobDataCSV
    from aperturedb.ConnectionDataCSV import ConnectionDataCSV
    from aperturedb.PolygonDataCSV import PolygonDataCSV
    from aperturedb.VideoDataCSV import VideoDataCSV
    from aperturedb.DescriptorDataCSV import DescriptorDataCSV
    from aperturedb.DescriptorSetDataCSV import DescriptorSetDataCSV

    ingest_types = {
        IngestType.BLOB: BlobDataCSV,
        IngestType.BOUNDING_BOX: BBoxDataCSV,
        IngestType.CONNECTION: ConnectionDataCSV,
        IngestType.DESCRIPTOR: DescriptorDataCSV,
        IngestType.DESCRIPTORSET: DescriptorSetDataCSV,
        IngestType.ENTITY: EntityDataCSV,
        IngestType.IMAGE: ImageDataCSV,
        IngestType.POLYGON: PolygonDataCSV,
        IngestType.VIDEO: VideoDataCSV
    }

    data = ingest_types[ingest_type](filepath, use_dask=use_dask,
                                     blobs_relative_to_csv=blobs_relative_to_csv)
    data.sample_count = len(data) if sample_count == -1 else sample_count
    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{ingest_type}.{os.path.basename(filepath)}")

    _process_data(
        data,
        sample_count=sample_count,
        module_name=os.path.basename(filepath),
        batchsize=batchsize,
        num_workers=num_workers,
        stats=stats,
        debug=debug
    )


@app.command()
def from_croissant(
    url: Annotated[str, typer.Argument(
        help="URL to the MLCroissant dataset")],
    flatten_json: Annotated[bool, typer.Option(
        help="Flatten JSON data")] = False,
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
    debug: Annotated[bool, typer.Option(
        help="Debug mode")] = False,
    stats: Annotated[bool, typer.Option(
        help="Show realtime statistics with summary")] = True,
    batchsize: Annotated[int, typer.Option(
        help="Size of the batch")] = 1,
    num_workers: Annotated[int, typer.Option(
        help="Number of workers for ingestion")] = 1,
):
    """
    Ingest data from MLCroissant dataset.
    """
    try:
        import mlcroissant as mlc
    except ImportError:
        console.log(
            "mlcroissant is not installed. Please install it with `pip install -U mlcroissant`.")
        typer.Abort()

    from aperturedb.MLCroissant import MLCroissantRecordSet, persist_metadata

    croissant_dataset = mlc.Dataset(url)
    metadata = persist_metadata(croissant_dataset, url)
    _process_data(
        [metadata],
        sample_count=1,
        module_name="croissant_metadata",
        batchsize=batchsize,
        num_workers=num_workers,
        stats=stats,
        debug=debug
    )
    for record_set in croissant_dataset.metadata.record_sets:
        console.log(
            f"Record Set: {record_set.name} ({record_set.uuid}) with {sample_count} samples")

        data = MLCroissantRecordSet(
            croissant_dataset.records(record_set=record_set.uuid),
            name=record_set.name or record_set.uuid,
            flatten_json=flatten_json,
            sample_count=sample_count,
            uuid=record_set.uuid
        )
        actual_sample_count = len(data)
        _process_data(
            data,
            sample_count=actual_sample_count,
            module_name=record_set.name or record_set.uuid,
            batchsize=batchsize,
            num_workers=num_workers,
            stats=stats,
            debug=debug
        )


@app.command()
def generate_embedding_csv_from_image_csv(
    input_file: Annotated[str, typer.Argument(
        help="Path to csv for ingestion")],
    set_name: Annotated[str, typer.Argument(
        help="Descriptor set name to be associated with the embeddings")],
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times.]")] = None,
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
):
    """
    Create a DescriptorDataCSV from an ImageDataCSV, using the specified transformer.
    Also save embeddings in a separate file, and generate the ConnectionDataCSV
    for linking the images to the descriptors.
    The ingestion of the embeddings and the connection data is not done here.
    image_properties transformer must be used at actual ingestion.
    It is also applied by default for CSV generation.
    """
    import pandas as pd
    from aperturedb.ImageDataCSV import ImageDataCSV
    import numpy as np

    data = ImageDataCSV(input_file)
    data.sample_count = len(data) if sample_count == -1 else sample_count
    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        if len(all_transformers) == 0:
            console.log(
                "No transformer specified . Generating embeddings from raw images requires a transformer.")
            typer.Abort()
        all_transformers.insert(0, "image_properties")
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{os.path.basename(input_file)}")
    filename = f"{os.path.basename(input_file)}_{all_transformers[-1]}"
    metadata = []
    connection = []
    embeddings = []
    errored = []

    for i in tqdm(range(data.sample_count)):
        d = data[i]
        nparr = np.frombuffer(d[1][1], dtype=np.float32)
        if not nparr.any():
            errored.append(d[0][0]["AddImage"]["properties"])

        embeddings.append(nparr)
        descriptor_id = f"{set_name}_{os.path.basename(input_file)}_{i}"
        metadata.append({
            "filename": f"{filename}.npy",
            "index": i,
            "set": set_name,
            "id": descriptor_id,
            "constraint_id": descriptor_id
        })
        connection.append({
            "ConnectionClass": "image_descriptor",
            "_Image@adb_image_sha256": d[0][0]["AddImage"]["properties"]["adb_image_sha256"],
            "_Descriptor@id": descriptor_id,
        })

    with open(f"{filename}.npy", "wb") as f:
        np.save(f, embeddings)
    pd.json_normalize(metadata).to_csv(
        f"{filename}_metadata.adb.csv", index=False)
    pd.json_normalize(connection).to_csv(
        f"{filename}_connection.adb.csv", index=False)
    if(len(errored) > 0):
        pd.json_normalize(errored).to_csv(
            f"{filename}_errored.csv", index=False)



from aperturedb import ParallelQuery
from aperturedb.Connector import Connector
from aperturedb.Utils import Utils
from aperturedb.Subscriptable import Subscriptable

import numpy as np
import logging

# For each property of each Entity or Connection,
# the following information is returned as an array of 3 elements of the form [matched, has_index_flag, type]:
# - matched: Number of objects that match the search.
# - has_index_flag: Indicates whether the property is indexed or not.
# - type: Type for the property. See supported types.
# https://docs.aperturedata.io/query_language/Reference/db_commands/GetSchema
PROPERTIES_SCHEMA_INDEX_FLAG = 1

logger = logging.getLogger(__name__)


class ParallelLoader(ParallelQuery.ParallelQuery):
    """
    **Parallel and Batch Loader for ApertureDB**

    This takes a dataset (which is a collection of homogeneous objects)
    or a derived class, and optimally inserts them into database by splitting them
    into batches, and passing the batches to multiple workers.
    """

    def __init__(self, client: Connector, dry_run: bool = False):
        super().__init__(client, dry_run=dry_run)
        self.utils = Utils(self.client)
        self.type = "element"

    def get_entity_indexes(self, schema: dict) -> dict:
        """
        Returns a dictionary of indexes for entities' properties.

        Args:
            schema (dict): The schema dictionary to get indexes from.

        Returns:
            dict: A dictionary of entity indexes.
        """

        indexes = {}

        entities = schema.get("entities") or {}

        for cls_name, cls_schema in (entities.get("classes") or {}).items():
            for prop_name, prop_schema in (cls_schema.get("properties") or {}).items():
                if prop_schema[PROPERTIES_SCHEMA_INDEX_FLAG]:
                    indexes.setdefault("entity", {}).setdefault(
                        cls_name, set()).add(prop_name)

        return indexes

    def get_connection_indexes(self, schema: dict) -> dict:
        """
        Returns a dictionary of indexes for connections' properties.

        Args:
            schema (dict): The schema dictionary to get indexes from.

        Returns:
            dict: A dictionary of connection indexes.
        """

        indexes = {}

        connections = schema.get("connections") or {}
        cls_names = connections.get("classes") or {}

        for cls_name, cls_schema in cls_names.items():

            # check if cls_schema is a dict or a list
            if isinstance(cls_schema, dict):
                for prop_name, prop_schema in (cls_schema.get("properties") or {}).items():
                    if prop_schema[PROPERTIES_SCHEMA_INDEX_FLAG]:
                        indexes.setdefault("connection", {}).setdefault(
                            cls_name, set()).add(prop_name)
            elif isinstance(cls_schema, list):
                # If cls_schema is a list, this occurs when the schema defines multiple connection variants
                # for the same class. Each element in the list is expected to be a dictionary representing
                # a connection variant, with a "properties" key containing the property schemas.
                # Example schema format:
                # "connections": {
                #     "classes": {
                #         "SomeConnectionClass": [
                #             {"properties": {"prop1": [...], "prop2": [...]}},
                #             {"properties": {"prop3": [...], "prop4": [...]}},
                #         ]
                #     }
                # }
                for connection in cls_schema:
                    for prop_name, prop_schema in (connection.get("properties") or {}).items():
                        if prop_schema[PROPERTIES_SCHEMA_INDEX_FLAG]:
                            indexes.setdefault("connection", {}).setdefault(
                                cls_name, set()).add(prop_name)
            else:
                exception_msg = "Unexpected schema format for connection class "
                exception_msg += f"'{cls_name}': {cls_schema}"
                logger.error(exception_msg)
                raise ValueError(exception_msg)

        return indexes

    def get_existing_indices(self):

        indexes = {}
        schema = self.utils.get_schema()

        if schema:
            entity_indexes     = self.get_entity_indexes(schema)
            connection_indexes = self.get_connection_indexes(schema)

            # Combine both entity and connection indexes
            indexes = {**entity_indexes, **connection_indexes}

        return indexes

    def query_setup(self, generator: Subscriptable) -> None:
        """
        Runs the setup for the loader, which includes creating indices.
        Currently, it only creates indices for the properties that are used for constraint.

        Will only run when the argument generator has a get_indices method that returns
        a dictionary of the form:

        ``` python
        {
            "entity": {
                "class_name": ["property_name"]
            },
        }

        or

        {
            "connection": {
                "class_name": ["property_name"]
            },
        }
        ```

        Args:
            generator (Subscriptable): The Subscriptable object that is being ingested
        """
        if hasattr(generator, "get_indices"):
            schema = self.utils.get_schema()

            indices_needed = generator.get_indices()

            # Create indexes for entities
            for entity_class in indices_needed.get("entity", {}):
                for property_name in indices_needed["entity"].get(entity_class, []):
                    if not self.utils.create_entity_index(entity_class, property_name):
                        logger.warning(
                            f"Failed to create index for {entity_class}.{property_name}")

            # Create indexes for connections
            for connection_class in indices_needed.get("connection", {}):
                for property_name in indices_needed["connection"].get(connection_class, []):
                    if not self.utils.create_connection_index(connection_class, property_name):
                        logger.warning(
                            f"Failed to create index for {connection_class}.{property_name}")

    def ingest(self, generator: Subscriptable, batchsize: int = 1, numthreads: int = 4, stats: bool = False) -> None:
        """
        **Method to ingest data into the database**

        Args:
            generator (Subscriptable): The list of data, or a class derived from [Subscriptable](/python_sdk/helpers/Subscriptable) to be ingested.
            batchsize (int, optional): The size of batch to be used. Defaults to 1.
            numthreads (int, optional): Number of workers to create. Defaults to 4.
            stats (bool, optional): If stats need to be presented, realtime. Defaults to False.
        """
        logger.info(
            f"Starting ingestion with batchsize={batchsize}, numthreads={numthreads}")
        self.query(generator, batchsize, numthreads, stats)

    def print_stats(self) -> None:

        times = np.array(self.get_times())
        total_queries_exec = len(times)

        print("============ ApertureDB Loader Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        succeeded_queries = sum([stat["succeeded_queries"]
                                for stat in self.actual_stats])
        succeeded_commands = sum([stat["succeeded_commands"]
                                  for stat in self.actual_stats])

        if succeeded_queries == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 0 if mean == 0 else 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

            # TODO this does not take into account that the last
            # batch may be smaller than batchsize
            print(f"Total inserted elements: {succeeded_queries}")
            print(f"Total successful commands: {succeeded_commands}")
        print("=================================================")



import dataclasses
import hashlib
import io
import json
import logging
import PIL.GifImagePlugin
import mlcroissant as mlc
import PIL.Image
import pandas as pd

from typing import Any, List, Tuple

from aperturedb.Subscriptable import Subscriptable
from aperturedb.Query import QueryBuilder
from aperturedb.DataModels import IdentityDataModel
from aperturedb.Query import generate_add_query


logger = logging.getLogger(__name__)


MAX_REF_VALUE = 99999
# This is useful to identify the class of the record in ApertureDB.
CLASS_PROPERTY_NAME = "adb_class_name"


class RecordSetModel(IdentityDataModel):
    name: str
    description: str = ""
    uuid: str = ""


class DatasetModel(IdentityDataModel):
    url: str = ""
    name: str = "Croissant Dataset automatically ingested into ApertureDB"
    description: str = f"A dataset loaded from a croissant json-ld"
    version: str = "1.0.0"
    record_sets: List[RecordSetModel] = dataclasses.field(default_factory=list)


def deserialize_record(record):
    """These are the types of records that we expect to deserialize, from croissant Records.

    Args:
        record (_type_): _description_

    Returns:
        _type_: _description_
    """
    deserialized = record
    if record is None:
        deserialized = "Not Available"
    if isinstance(record, bytes):
        deserialized = record.decode('utf-8')
    if isinstance(record, pd.Timestamp):
        deserialized = {"_date": record.to_pydatetime().isoformat()}
    if record == pd.NaT:
        deserialized = "Not Available Time"
    if isinstance(deserialized, str):
        if deserialized.startswith("[") or deserialized.startswith("{"):
            # If it looks like a list or dict, try to parse it as JSON
            try:
                deserialized = json.loads(deserialized)
            except json.JSONDecodeError:
                logger.info(f"Failed to parse JSON: {deserialized}")

            try:
                deserialized = json.loads(deserialized.replace("'", "\""))
            except Exception as e:
                logger.info(
                    f"Failed to parse JSON: {deserialized} with error {e}")

    if isinstance(deserialized, list):
        deserialized = [deserialize_record(item) for item in deserialized]
    if isinstance(deserialized, dict):
        deserialized = {k: deserialize_record(
            v) for k, v in deserialized.items()}

    return deserialized


def persist_metadata(dataset: mlc.Dataset, url: str) -> Tuple[List[dict], List[bytes]]:
    """
    Persist the metadata of a croissant dataset into ApertureDB.
    """
    ds = DatasetModel(
        url=url,
        name=dataset.metadata.name,
        description=dataset.metadata.description,
        version=dataset.metadata.version or "1.0.0",
        record_sets=[RecordSetModel(
            name=rs.name,
            description=rs.description,
            uuid=rs.uuid,
        ) for rs in dataset.metadata.record_sets]
    )
    q, b, _ = generate_add_query(ds)

    return q, b


def try_parse(value: str) -> Any:
    """Attempts to parse a string value into a more appropriate type."""
    parsed = value.strip()

    if parsed.startswith("http"):
        # Download the content from the URL
        from aperturedb.Sources import Sources
        sources = Sources(n_download_retries=3)
        result, buffer = sources.load_from_http_url(
            parsed, validator=lambda x: True)
        if result:
            parsed = PIL.Image.open(io.BytesIO(buffer))

    return parsed


def dict_to_query(row_dict, name: str, flatten_json: bool) -> Any:
    literals = {}
    subitems = {}
    known_image_blobs = {}
    unknown_blobs = {}
    o_literals = {}

    name = name.split("/")[-1]  # Use the last part of the name
    # If name is not specified, or begins with _, this ensures that it
    # complies with the ApertureDB naming conventions
    if not name or name.startswith("_"):
        safe_name = f"E_{name or 'Record'}"  # Uncomment if you want
        logger.warning(
            f"Entity Name '{name}' is not valid. Using {safe_name}.")
        name = safe_name

    for k, v in row_dict.items():
        k = k.split("/")[-1]  # Use the last part of the key
        if not k or k.startswith("_"):
            safe_key = f"F_{k or 'Field'}"
            logger.warning(
                f"Property name '{k}' is not valid. Using {safe_key}.")
            k = safe_key
        item = v
        # Pre processed items from croissant.
        if isinstance(item, PIL.Image.Image):
            buffer = io.BytesIO()
            item.save(buffer, format=item.format)
            known_image_blobs[k] = buffer.getvalue()
            continue

        record = deserialize_record(item)
        if isinstance(record, str):
            record = try_parse(record)

            # Post processed items from SDK.
            if isinstance(record, PIL.GifImagePlugin.GifImageFile):
                buffer = io.BytesIO()
                record.save(buffer, format=record.format)
                unknown_blobs[k] = buffer.getvalue()
                continue

            if isinstance(record, PIL.Image.Image):
                buffer = io.BytesIO()
                record.save(buffer, format=record.format)
                known_image_blobs[k] = buffer.getvalue()
                continue

        if flatten_json and isinstance(record, list):
            subitems[k] = record
        else:
            literals[k] = record
            # Original value from croissant. This is useful for debugging.
            o_literals[k] = item

    if flatten_json:
        str_rep = "".join([f"{str(k)}{str(v)}" for k, v in literals.items()])
        literals["adb_uuid"] = hashlib.sha256(
            str_rep.encode('utf-8')).hexdigest()

    literals[CLASS_PROPERTY_NAME] = name
    q = QueryBuilder.add_command(name, {
        "properties": literals,
        "connect": {
            "ref": MAX_REF_VALUE,
            "class": "hasRecord",
            "direction": "in",
        }
    })
    if flatten_json:
        q[list(q.keys())[-1]]["if_not_found"] = {
            "adb_uuid": ["==", literals["adb_uuid"]]
        }

    dependents = []
    if len(subitems) > 0 or len(known_image_blobs) > 0 or len(unknown_blobs) > 0:
        # We need to create a reference to this record
        q[list(q.keys())[-1]]["_ref"] = 1

    for key in subitems:
        for item in subitems[key]:
            subitem_query, blobs = dict_to_query(
                item, f"{name}.{key}", flatten_json)
            subitem_query[0][list(subitem_query[0].keys())[-1]]["connect"] = {
                "ref": 1,
                "class": key,
                "direction": "in",
            }
            dependents.extend(subitem_query)

    from aperturedb.Query import ObjectType
    blobs = []
    for blob in known_image_blobs:
        image_query = QueryBuilder.add_command(ObjectType.IMAGE, {
            "properties": {CLASS_PROPERTY_NAME: literals[CLASS_PROPERTY_NAME] + "." + "image"},
            "connect": {
                "ref": 1,
                "class": blob,
                "direction": "in"
            }
        })
        blobs.append(known_image_blobs[blob])
        dependents.append(image_query)

    for blob in unknown_blobs:
        blob_query = QueryBuilder.add_command(ObjectType.BLOB, {
            "properties": {CLASS_PROPERTY_NAME: literals[CLASS_PROPERTY_NAME] + "." + "blob"},
            "connect": {
                "ref": 1,
                "class": blob,
                "direction": "in"
            }
        })
        blobs.append(unknown_blobs[blob])
        dependents.append(blob_query)

    return [q] + dependents, blobs


class MLCroissantRecordSet(Subscriptable):
    def __init__(
            self,
            record_set: mlc.Records,
            name: str,
            flatten_json: bool,
            sample_count: int = 0,
            uuid: str = None):
        self.record_set = record_set
        self.uuid = uuid
        samples = []
        count = 0
        for record in record_set:
            samples.append({k: v for k, v in record.items()})
            count += 1
            if count == sample_count:
                break

        self.samples = samples
        self.sample_count = len(samples)
        self.name = name
        self.flatten_json = flatten_json
        self.indexed_entities = set()

    def getitem(self, subscript):
        row_dict = self.samples[subscript]

        find_recordset_query = QueryBuilder.find_command(
            "RecordSetModel", {
                "_ref": MAX_REF_VALUE,
                "constraints": {
                    "uuid": ["==", self.uuid]
                }
            })

        q, blobs = dict_to_query(row_dict, self.name, self.flatten_json)
        indexes_to_create = []
        for command in q:
            cmd = list(command.keys())[-1]
            if cmd in ["AddImage", "AddBlob", "AddVideo"]:
                continue
            indexable_entity = command[list(command.keys())[-1]]["class"]
            if indexable_entity not in self.indexed_entities:
                index_command = {
                    "CreateIndex": {
                        "class": indexable_entity,
                        "index_type": "entity",
                        "property_key": "adb_uuid",
                    }
                }
                indexes_to_create.append(index_command)
        return indexes_to_create + [find_recordset_query] + q, blobs

    def __len__(self):
        return len(self.samples)
